sound_type snd_make_stkchorus(sound_type s1, double baseDelay, double depth, double freq, double mix);
sound_type snd_stkchorus(sound_type s1, double baseDelay, double depth, double freq, double mix);
    /* LISP: (snd-stkchorus SOUND ANYNUM ANYNUM ANYNUM ANYNUM) */
