/**********************************************************************

  Audacity: A Digital Audio Editor

  examples.hh

  James Crook

********************************************************************//*

JKC: This file is for the doxygen documentation system.
We can use it for small example that have '/ *' style comments in 
them, which cannot be included in Doxy documents otherwise.

/*****************************************************************//*! 

//First Example:

   /* i18n-hint: One-letter abbreviation for Left, in VU Meter */
   mLeftText = _("L");
   /* i18n-hint: One-letter abbreviation for Right, in VU Meter */
   mRightText = _("R");
