extern LVAL xlc_seq_reset(void);
extern LVAL xlc_seq_insert_ctrl(void);
extern LVAL xlc_seq_insert_ramp(void);
extern LVAL xlc_seq_insert_macctrl(void);
extern LVAL xlc_seq_insert_note(void);
extern LVAL xlc_seq_copy(void);
extern LVAL xlc_seq_create(void);
extern LVAL xlc_seq_next(void);
extern LVAL xlc_seq_get(void);
extern LVAL xlc_seq_write_smf(void);
extern LVAL xlc_seq_read_smf(void);
extern LVAL xlc_seq_read(void);
extern LVAL xlc_seq_write(void);
