/*
* Audacity: A Digital Audio Editor
*/
#include "recordconfiguration.h"

using namespace muse;
using namespace au::record;

draw::Color RecordConfiguration::recordColor() const
{
    return "#EF476F";
}
