mv path.data $1.path.data
mv smooth.data $1.smooth.data
mv beatmap.txt $1.beatmap.txt
mv debug-log.txt $1.debug-log.txt
mv midi.mid $1.midi.mid
mv transcription.txt $1.transcription.txt
echo "moved outputs to backup files prefixed with $1."

 