sound_type snd_make_stkrev(int rev_type, sound_type s1, time_type trev, double mix);
sound_type snd_stkrev(int rev_type, sound_type s1, time_type trev, double mix);
    /* LISP: (snd-stkrev LONG SOUND ANYNUM ANYNUM) */
