nyqsrc/sndfnint nylsf/sndfile.h nyqsrc/sndfmt.h
	nyqsrc/sound.h nyqsrc/add.h   
	nyqsrc/avg.h nyqsrc/compose.h   
	nyqsrc/convolve.h nyqsrc/downsample.h   
	nyqsrc/fft.h nyqsrc/inverse.h   
	nyqsrc/multiseq.h nyqsrc/resamp.h   
	nyqsrc/resampv.h nyqsrc/samples.h   
	nyqsrc/sndmax.h nyqsrc/sndread.h   
	nyqsrc/sndseq.h nyqsrc/sndsliders.h   
	nyqsrc/sndwrite.h nyqsrc/yin.h   
	nyqsrc/nyq-osc-server.h nyqsrc/trigger.h   
	nyqsrc/lpanal.h nyqsrc/phasevocoder.h   
	nyqsrc/pvshell.h 
	tran/abs.h tran/allpoles.h   
	tran/alpass.h tran/alpasscv.h   
	tran/alpassvv.h tran/amosc.h   
	tran/areson.h tran/aresonvc.h   
	tran/aresoncv.h tran/aresonvv.h   
	tran/atone.h tran/atonev.h   
	tran/biquadfilt.h tran/buzz.h   
	tran/chase.h tran/clip.h   
	tran/congen.h tran/const.h   
	tran/coterm.h tran/delaycc.h   
	tran/delaycv.h tran/eqbandvvv.h   
	tran/exp.h tran/follow.h   
	tran/fmosc.h tran/fromobject.h   
	tran/fromarraystream.h tran/gate.h   
	tran/ifft.h tran/instrclar.h   
	tran/instrclarall.h tran/instrclarfreq.h   
	tran/instrsax.h tran/instrsaxall.h   
	tran/instrsaxfreq.h tran/integrate.h   
	tran/log.h tran/lpreson.h   
	tran/maxv.h tran/offset.h   
	tran/oneshot.h tran/osc.h   
	tran/partial.h tran/pluck.h   
	tran/prod.h tran/pwl.h   
	tran/quantize.h tran/recip.h   
	tran/reson.h tran/resonvc.h   
	tran/resoncv.h tran/resonvv.h   
	tran/sampler.h tran/scale.h   
	tran/shape.h tran/sine.h   
	tran/siosc.h tran/slope.h   
	tran/sqrt.h tran/tapf.h   
	tran/tapv.h tran/tone.h   
	tran/tonev.h tran/upsample.h   
	tran/white.h tran/stkrev.h   
	tran/stkpitshift.h tran/stkchorus.h   
	tran/instrbow.h tran/instrbowedfreq.h   
	tran/instrbanded.h tran/instrmandolin.h   
	tran/instrsitar.h tran/instrmodalbar.h   
	tran/instrflute.h tran/instrflutefreq.h   
	tran/instrfluteall.h tran/fmfb.h   
	tran/fmfbv.h 

