/**********************************************************************

  Audacity: A Digital Audio Editor

  AudioIOSequences.cpp

  Paul Licameli

**********************************************************************/
#include "AudioIOSequences.h"

PlayableSequence::~PlayableSequence() = default;

RecordableSequence::~RecordableSequence() = default;

OtherPlayableSequence::~OtherPlayableSequence() = default;
