/**********************************************************************
 
 Audacity: A Digital Audio Editor
 
 @file wxArrayStringEx.cpp
 
 Paul Licameli split from MemoryX.h
 
 **********************************************************************/

// This .cpp exists just to ensure there is at least one that includes
// the header before any other header, and compiles
#include "wxArrayStringEx.h"
