AudacityLogo ICON "audacity.ico"

#include <winuser.h>
CREATEPROCESS_MANIFEST_RESOURCE_ID RT_MANIFEST "Audacity.exe.manifest"

#include "wx/msw/wx.rc"
/////////////////////////////////////////////////////////////////////////////
//
// Version
//
VS_VERSION_INFO VERSIONINFO
 FILEVERSION AUDACITY_VERSION,AUDACITY_RELEASE,AUDACITY_REVISION,0
 PRODUCTVERSION AUDACITY_VERSION,AUDACITY_RELEASE,AUDACITY_REVISION,0
 FILEFLAGSMASK 0x17L
#ifdef _DEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS 0x4L
 FILETYPE 0x1L
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "CompanyName", "Audacity Team"
            VALUE "FileDescription", "Audacity"
            VALUE "FileVersion", AUDACITY_FILE_VERSION
            VALUE "InternalName", "Audacity"
            VALUE "LegalCopyright", "Copyright © 2024. All rights reserved."
            VALUE "LegalTrademarks", "Audacity® is a registered trademark of Muse Group."
            VALUE "OriginalFilename", "Audacity.exe"
            VALUE "ProductName", "Audacity"
            VALUE "ProductVersion", AUDACITY_FILE_VERSION
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END
