/**********************************************************************

  Audacity: A Digital Audio Editor

  MusicalInstruments.h

  Vaughan Johnson, January 2007

  Includes musical instrument images for use in MixerTrackCluster.

**********************************************************************/

#include "MusicalInstruments/acoustic_guitar_gtr.xpm"
#include "MusicalInstruments/acoustic_piano_pno.xpm"
#include "MusicalInstruments/back_vocal_bg_vox.xpm"
#include "MusicalInstruments/clap.xpm"

#include "MusicalInstruments/drums_dr.xpm"
#include "MusicalInstruments/electric_bass_guitar_bs_gtr.xpm"
#include "MusicalInstruments/electric_guitar_gtr.xpm"
#include "MusicalInstruments/electric_piano_pno_key.xpm"

#include "MusicalInstruments/kick.xpm"
#include "MusicalInstruments/loop.xpm"
#include "MusicalInstruments/organ_org.xpm"
#include "MusicalInstruments/perc.xpm"

#include "MusicalInstruments/sax.xpm"
#include "MusicalInstruments/snare.xpm"
#include "MusicalInstruments/string_violin_cello.xpm"
#include "MusicalInstruments/synth.xpm"

#include "MusicalInstruments/tambo.xpm"
#include "MusicalInstruments/trumpet_horn.xpm"
#include "MusicalInstruments/turntable.xpm"
#include "MusicalInstruments/vibraphone_vibes.xpm"

#include "MusicalInstruments/vocal_vox.xpm"

// This one should always be last, so it wins when best score is 0.
#include "MusicalInstruments/_default_instrument.xpm"
