#include "down-9x16.xpm"
#include "fav-down-16x16.xpm"
#include "fav-up-16x16.xpm"
#include "power-off-16x16.xpm"
#include "power-on-16x16.xpm"
#include "remove-16x16.xpm"
#include "settings-down-16x16.xpm"
#include "settings-up-16x16.xpm"
#include "up-9x16.xpm"
