/*
* Audacity: A Digital Audio Editor
*/
#include "nyquisteffectsrepository.h"

using namespace au::effects;

EffectMetaList NyquistEffectsRepository::effectMetaList() const
{
    EffectMetaList effects;

    //! TODO This is just blank

    return effects;
}
