extern void midifile(void);
extern int (*Mf_getc)(void);
extern void (*Mf_header)(int,int,int);
extern void (*Mf_starttrack)(void);
extern void (*Mf_endtrack)(void);
extern void (*Mf_on)(int,int,int);
extern void (*Mf_off)(int,int,int);
extern void (*Mf_pressure)(int,int,int);
extern void (*Mf_controller)(int,int,int);
extern void (*Mf_pitchbend)(int,int,int);
extern void (*Mf_program)(int,int);
extern void (*Mf_chanpressure)(int,int);
extern void (*Mf_sysex)(int,char*);
extern void (*Mf_metamisc)(int,int,char*);
extern void (*Mf_sqspecific)(int,char*);
extern void (*Mf_seqnum)(int);
extern void (*Mf_text)(int,int,char*);
extern void (*Mf_eot)(void);
extern void (*Mf_timesig)(int,int,int,int);
extern void (*Mf_smpte)(int,int,int,int,int);
extern void (*Mf_tempo)(int);
extern void (*Mf_keysig)(int,int);
extern void (*Mf_arbitrary)(int,char*);
extern void (*Mf_error)(char *);
extern long Mf_currtime;
extern int Mf_nomerge;
extern int Mf_skipinit;
