#! /usr/bin/env bash

export APPIMAGE_PRESERVED_GTK_DATA_PREFIX=${GTK_DATA_PREFIX}
export APPIMAGE_PRESERVED_GTK_THEME=${GTK_THEME}
export APPIMAGE_PRESERVED_GDK_BACKEND=${GDK_BACKEND}
export APPIMAGE_PRESERVED_XDG_DATA_DIRS=${XDG_DATA_DIRS}
export APPIMAGE_PRESERVED_GSETTINGS_SCHEMA_DIR=${GSETTINGS_SCHEMA_DIR}
export APPIMAGE_PRESERVED_GI_TYPELIB_PATH=${GI_TYPELIB_PATH}
export APPIMAGE_PRESERVED_GDK_PIXBUF_MODULE_FILE=${GDK_PIXBUF_MODULE_FILE}
export APPIMAGE_PRESERVED_LD_LIBRARY_PATH=${LD_LIBRARY_PATH}
export APPIMAGE_PRESERVED_GIO_EXTRA_MODULES=${GIO_EXTRA_MODULES}
