/*
* Audacity: A Digital Audio Editor
*/
#include "playbackconfiguration.h"

using namespace muse;
using namespace au::playback;

draw::Color PlaybackConfiguration::playColor() const
{
    return "#18A999";
}
