sound_type snd_make_mandolin(time_type t0, double freq, time_type d, double body_size, double detune, rate_type sr);
sound_type snd_mandolin(time_type t0, double freq, time_type d, double body_size, double detune, rate_type sr);
    /* LISP: (snd-mandolin ANYNUM ANYNUM ANYNUM ANYNUM ANYNUM ANYNUM) */
#define MAND_CONTROL_CHANGE_CONST 128
