/**********************************************************************

  Audacity: A Digital Audio Editor

  LoadThemeResources.cpp

  Paul Licameli

**********************************************************************/

#ifndef __AUDACITY_LOAD_THEME_RESOURCES__
#define __AUDACITY_LOAD_THEME_RESOURCES__

namespace ThemeResources {
   THEME_RESOURCES_API void Load();
}

#endif
