sound_type snd_make_clarinet_freq(double freq1, sound_type breath_env, sound_type freq_env, rate_type sr);
sound_type snd_clarinet_freq(double freq1, sound_type breath_env, sound_type freq_env, rate_type sr);
    /* LISP: (snd-clarinet_freq ANYNUM SOUND SOUND ANYNUM) */
#define CLAR_CONTROL_CHANGE_CONST 128
