  { "SND-SET-MAX-AUDIO-MEM",  S, xlc_snd_set_max_audio_mem}, 
  { "SND-SET-LATENCY",  S, xlc_snd_set_latency}, 
  { "SOUNDP",  S, xlc_soundp}, 
  { "HZ-TO-STEP",  S, xlc_hz_to_step}, 
  { "SND-SET-LOGICAL-STOP",  S, xlc_snd_set_logical_stop}, 
  { "LOG",  S, xlc_log}, 
  { "SND-SREF",  S, xlc_snd_sref}, 
  { "SREF-INVERSE",  S, xlc_sref_inverse}, 
  { "SND-STOP-TIME",  S, xlc_snd_stop_time}, 
  { "SND-TIME",  S, xlc_snd_time}, 
  { "SND-SRATE",  S, xlc_snd_srate}, 
  { "SND-T0",  S, xlc_snd_t0}, 
  { "SND-XFORM",  S, xlc_snd_xform}, 
  { "BLOCK-WATCH",  S, xlc_block_watch}, 
  { "SOUND-NTH-BLOCK",  S, xlc_sound_nth_block}, 
  { "SND-COPY",  S, xlc_snd_copy}, 
  { "SND-PRINT",  S, xlc_snd_print}, 
  { "SND-PLAY",  S, xlc_snd_play}, 
  { "STATS",  S, xlc_stats}, 
  { "SND-PRINT-TREE",  S, xlc_snd_print_tree}, 
  { "SND-SCALE",  S, xlc_snd_scale}, 
  { "SND-ZERO",  S, xlc_snd_zero}, 
  { "STEP-TO-HZ",  S, xlc_step_to_hz}, 
  { "SND-ADD",  S, xlc_snd_add}, 
  { "SND-AVG",  S, xlc_snd_avg}, 
  { "SND-COMPOSE",  S, xlc_snd_compose}, 
  { "SND-CONVOLVE",  S, xlc_snd_convolve}, 
  { "SND-DOWN",  S, xlc_snd_down}, 
  { "SND-FFT",  S, xlc_snd_fft}, 
  { "SND-INVERSE",  S, xlc_snd_inverse}, 
  { "SND-LPANAL",  S, xlc_snd_lpanal}, 
  { "SND-MULTISEQ",  S, xlc_snd_multiseq}, 
  { "SND-PHASEVOCODER",  S, xlc_snd_phasevocoder}, 
  { "SND-RESAMPLE",  S, xlc_snd_resample}, 
  { "SND-RESAMPLEV",  S, xlc_snd_resamplev}, 
  { "SND-FROM-ARRAY",  S, xlc_snd_from_array}, 
  { "SND-SAMPLES",  S, xlc_snd_samples}, 
  { "SND-LENGTH",  S, xlc_snd_length}, 
  { "SND-MAXSAMP",  S, xlc_snd_maxsamp}, 
  { "SND-FETCH",  S, xlc_snd_fetch}, 
  { "SND-FETCH-ARRAY",  S, xlc_snd_fetch_array}, 
  { "SND-MAX",  S, xlc_snd_max}, 
  { "SND-READ",  S, xlc_snd_read}, 
  { "SND-SEQ",  S, xlc_snd_seq}, 
  { "SND-SLIDER",  S, xlc_snd_slider}, 
  { "SND-STOPONZERO",  S, xlc_snd_stoponzero}, 
  { "SND-TRIGGER",  S, xlc_snd_trigger}, 
  { "SND-YIN",  S, xlc_snd_yin}, 
  { "SND-ABS",  S, xlc_snd_abs}, 
  { "SND-ALLPOLES",  S, xlc_snd_allpoles}, 
  { "SND-ALPASS",  S, xlc_snd_alpass}, 
  { "SND-ALPASSCV",  S, xlc_snd_alpasscv}, 
  { "SND-ALPASSVC",  S, xlc_snd_alpassvc}, 
  { "SND-ALPASSVV",  S, xlc_snd_alpassvv}, 
  { "SND-AMOSC",  S, xlc_snd_amosc}, 
  { "SND-ARESON",  S, xlc_snd_areson}, 
  { "SND-ARESONCV",  S, xlc_snd_aresoncv}, 
  { "SND-ARESONVC",  S, xlc_snd_aresonvc}, 
  { "SND-ARESONVV",  S, xlc_snd_aresonvv}, 
  { "SND-ATONE",  S, xlc_snd_atone}, 
  { "SND-ATONEV",  S, xlc_snd_atonev}, 
  { "SND-BIQUAD",  S, xlc_snd_biquad}, 
  { "SND-BUZZ",  S, xlc_snd_buzz}, 
  { "SND-CHASE",  S, xlc_snd_chase}, 
  { "SND-CLIP",  S, xlc_snd_clip}, 
  { "SND-CONGEN",  S, xlc_snd_congen}, 
  { "SND-CONST",  S, xlc_snd_const}, 
  { "SND-COTERM",  S, xlc_snd_coterm}, 
  { "SND-DELAY",  S, xlc_snd_delay}, 
  { "SND-DELAYCV",  S, xlc_snd_delaycv}, 
  { "SND-EQBANDVVV",  S, xlc_snd_eqbandvvv}, 
  { "SND-EXP",  S, xlc_snd_exp}, 
  { "SND-FMFB",  S, xlc_snd_fmfb}, 
  { "SND-FMFBV",  S, xlc_snd_fmfbv}, 
  { "SND-FMOSC",  S, xlc_snd_fmosc}, 
  { "SND-FOLLOW",  S, xlc_snd_follow}, 
  { "SND-FROMARRAYSTREAM",  S, xlc_snd_fromarraystream}, 
  { "SND-FROMOBJECT",  S, xlc_snd_fromobject}, 
  { "SND-GATE",  S, xlc_snd_gate}, 
  { "SND-IFFT",  S, xlc_snd_ifft}, 
  { "SND-BANDEDWG",  S, xlc_snd_bandedwg}, 
  { "SND-BOWED",  S, xlc_snd_bowed}, 
  { "SND-BOWED_FREQ",  S, xlc_snd_bowed_freq}, 
  { "SND-CLARINET",  S, xlc_snd_clarinet}, 
  { "SND-CLARINET_ALL",  S, xlc_snd_clarinet_all}, 
  { "SND-CLARINET_FREQ",  S, xlc_snd_clarinet_freq}, 
  { "SND-FLUTE",  S, xlc_snd_flute}, 
  { "SND-FLUTE_ALL",  S, xlc_snd_flute_all}, 
  { "SND-FLUTE_FREQ",  S, xlc_snd_flute_freq}, 
  { "SND-MANDOLIN",  S, xlc_snd_mandolin}, 
  { "SND-MODALBAR",  S, xlc_snd_modalbar}, 
  { "SND-SAX",  S, xlc_snd_sax}, 
  { "SND-SAX_ALL",  S, xlc_snd_sax_all}, 
  { "SND-SAX_FREQ",  S, xlc_snd_sax_freq}, 
  { "SND-SITAR",  S, xlc_snd_sitar}, 
  { "SND-INTEGRATE",  S, xlc_snd_integrate}, 
  { "SND-LOG",  S, xlc_snd_log}, 
  { "SND-LPRESON",  S, xlc_snd_lpreson}, 
  { "SND-MAXV",  S, xlc_snd_maxv}, 
  { "SND-OFFSET",  S, xlc_snd_offset}, 
  { "SND-ONESHOT",  S, xlc_snd_oneshot}, 
  { "SND-OSC",  S, xlc_snd_osc}, 
  { "SND-PARTIAL",  S, xlc_snd_partial}, 
  { "SND-PLUCK",  S, xlc_snd_pluck}, 
  { "SND-PROD",  S, xlc_snd_prod}, 
  { "SND-PWL",  S, xlc_snd_pwl}, 
  { "SND-QUANTIZE",  S, xlc_snd_quantize}, 
  { "SND-RECIP",  S, xlc_snd_recip}, 
  { "SND-RESON",  S, xlc_snd_reson}, 
  { "SND-RESONCV",  S, xlc_snd_resoncv}, 
  { "SND-RESONVC",  S, xlc_snd_resonvc}, 
  { "SND-RESONVV",  S, xlc_snd_resonvv}, 
  { "SND-SAMPLER",  S, xlc_snd_sampler}, 
  { "SND-NORMALIZE",  S, xlc_snd_normalize}, 
  { "SND-SHAPE",  S, xlc_snd_shape}, 
  { "SND-SINE",  S, xlc_snd_sine}, 
  { "SND-SIOSC",  S, xlc_snd_siosc}, 
  { "SND-SLOPE",  S, xlc_snd_slope}, 
  { "SND-SQRT",  S, xlc_snd_sqrt}, 
  { "SND-TAPF",  S, xlc_snd_tapf}, 
  { "SND-TAPV",  S, xlc_snd_tapv}, 
  { "SND-TONE",  S, xlc_snd_tone}, 
  { "SND-TONEV",  S, xlc_snd_tonev}, 
  { "SND-UP",  S, xlc_snd_up}, 
  { "SND-WHITE",  S, xlc_snd_white}, 
  { "SND-STKPITSHIFT",  S, xlc_snd_stkpitshift}, 
  { "SND-STKREV",  S, xlc_snd_stkrev}, 
  { "SND-STKCHORUS",  S, xlc_snd_stkchorus}, 
  { "SND-SAVE",  S, xlc_snd_save}, 
  { "SND-OVERWRITE",  S, xlc_snd_overwrite}, 
