sound_type snd_make_resonvc(sound_type s1, sound_type hz, double bw, int normalization);
sound_type snd_resonvc(sound_type s1, sound_type hz, double bw, int normalization);
    /* LISP: (snd-resonvc SOUND SOUND ANYNUM LONG) */
