// sndwritepa.h -- headers for sndwritepa.c
//
// Roger Dannenberg
// Jan 2013
//
// Since sndwritepa.c is a PortAudio based implementation of the sndwrite.h
// interface, we just include sndwrite.h (By doing this, we regularize the 
// system so that there is a .h file corresponding to each .c file.)

#include "sndwrite.h"
