/*
* Audacity: A Digital Audio Editor
*/
#include <gtest/gtest.h>

#include "log.h"

using namespace muse;

class ProjectScene_SampleTests : public ::testing::Test
{
public:
};

TEST_F(ProjectScene_SampleTests, Sample)
{
}
