sound_type snd_make_sax_freq(double freq, sound_type breath_env, sound_type freq_env, rate_type sr);
sound_type snd_sax_freq(double freq, sound_type breath_env, sound_type freq_env, rate_type sr);
    /* LISP: (snd-sax_freq ANYNUM SOUND SOUND ANYNUM) */
#define SAX_CONTROL_CHANGE_CONST 128
