/* This .h file declares everything you need to include seq.h */
/*
 * This file gets included by an interface to seq generated by intgen
 */
#include "switches.h"
#include "stdio.h"
#include "cext.h"
#include "userio.h"
#include "midifns.h"
#include "timebase.h"
#include "moxc.h"
