/* localptrs.h -- extend XLISP with these functions
 *
 * CHANGE LOG
 * 28-Apr-03  rbd  Removed "include switches.h" -- already included
 */
 
/* extension to xlisp */
#include "sndfnintptrs.h"
#include "seqfnintptrs.h"
