sound_type snd_make_flute_freq(double freq, sound_type breath_env, sound_type freq_env, rate_type sr);
sound_type snd_flute_freq(double freq, sound_type breath_env, sound_type freq_env, rate_type sr);
    /* LISP: (snd-flute_freq ANYNUM SOUND SOUND ANYNUM) */
#define FLUTE_CONTROL_CHANGE_CONST 128
