  { "SEQ-RESET",  S, xlc_seq_reset}, 
  { "SEQ-INSERT-CTRL",  S, xlc_seq_insert_ctrl}, 
  { "SEQ-INSERT-RAMP",  S, xlc_seq_insert_ramp}, 
  { "SEQ-INSERT-MACCTRL",  S, xlc_seq_insert_macctrl}, 
  { "SEQ-INSERT-NOTE",  S, xlc_seq_insert_note}, 
  { "SEQ-COPY",  S, xlc_seq_copy}, 
  { "SEQ-CREATE",  S, xlc_seq_create}, 
  { "SEQ-NEXT",  S, xlc_seq_next}, 
  { "SEQ-GET",  S, xlc_seq_get}, 
  { "SEQ-WRITE-SMF",  S, xlc_seq_write_smf}, 
  { "SEQ-READ-SMF",  S, xlc_seq_read_smf}, 
  { "SEQ-READ",  S, xlc_seq_read}, 
  { "SEQ-WRITE",  S, xlc_seq_write}, 
