/* phasevocoder.h -- this is a stub showing how you might hook a
   phase vocoder into Nyquist using pvshell 
 */

sound_type snd_phasevocoder(sound_type f, sound_type g, long fftsize, long hopsize, long mode);
    /* LISP: (snd-phasevocoder SOUND SOUND LONG LONG LONG) */
