/**********************************************************************

  Audacity: A Digital Audio Editor

  @file IPCChannel.cpp

  @author Vitaly Sverchinsky

  Part of lib-ipc library

**********************************************************************/

#include "IPCChannel.h"

IPCChannelStatusCallback::~IPCChannelStatusCallback() = default;
IPCChannel::~IPCChannel() = default;
